;; DCO calibration routine
;;
;; Copyright 2015 - By Michael Kohn
;; http://www.mikekohn.net/
;; mike@mikekohn.net
;;
;; Calibrate a DCO oscillator with a 32.768KHz crystal

.msp430
.include "msp430x2xx.inc"

RAM equ 0x0200
CPU_FREQ_LO equ RAM+0
CPU_FREQ_HI equ RAM+2
DCO_COUNT equ RAM+4
BAUD_DIV equ RAM+6

;  r4 = temp
;  r5 = DCO val
;  r6 = RSEL val
;  r7 = temp
;  r8 = temp
;  r9 = Seconds counter
; r10 = DCO interrupt count
; r11 = Previous DCO interrupt count
; r12 =
; r13 = temp in calibrate
; r14 = temp
; r15 = temp

  .org 0xc000
start:
  ;; Turn off watchdog
  mov.w #(WDTPW|WDTHOLD), &WDTCTL

  ;; Turn off interrupts
  dint

  ;; Set up stack pointer
  mov.w #0x0400, SP

  ;; Set MCLK to 16 MHz with DCO 
  mov.b #DCO_4, &DCOCTL
  mov.b #RSEL_15, &BCSCTL1
  mov.b #0, &BCSCTL2

  ;; Set SMCLK to 32.768kHz external crystal
  mov.b #XCAP_3, &BCSCTL3

  ;; Set up output pins
  ;; P1.1 = RX
  ;; P1.2 = TX
  ;; P1.3 = Green LED
  ;; P1.4 = Red LED
  mov.b #0x1c, &P1DIR
  mov.b #0x1c, &P1OUT
  mov.b #0x06, &P1SEL
  mov.b #0x06, &P1SEL2

  ;; Set up Timer A
  mov.w #512, &TACCR0
  mov.w #TASSEL_2|MC_1, &TACTL ; SMCLK, DIV1, COUNT to TACCR0
  mov.w #CCIE, &TACCTL0
  mov.w #0, &TACCTL1

  ;; Set up Timer B
  mov.w #32768, &TBCCR0  ; 32768 ticks = 1 second
  mov.w #TBSSEL_1|MC_1, &TBCTL ; ACLK, DIV1, COUNT to TBCCR0
  mov.w #CCIE, &TBCCTL0
  mov.w #0, &TBCCTL1

  ;; Clear seconds counter and dco calibration registers
  mov.w #0, r9
  mov.w #0, r10
  mov.w #0, r11

  ;; Set RSEL, DCO
  mov.w #3, r5
  mov.w #15, r6

  ;; Turn on interrupts
  eint

  call #wait_on_dco_sample
  call #calibrate_baud_rate
  call #send_settings

main:
  mov.b #15, r6
next_rsel:
  mov.b #0, r5
next_dco:
  call #set_dco
  call #wait_on_dco_sample
  call #calibrate_baud_rate
  call #send_settings
  add.b #1, r5
  cmp.b #8, r5
  jne next_dco
  sub.b #1, r6
  cmp.b #0, r6
  jne next_rsel

  mov.w #done_string, r15
  call #send_string
while_1:
  jmp while_1

set_dco:
  mov.b &DCOCTL, r7
  mov.b &BCSCTL1, r8
  bic.b #0xe0, r7
  bic.b #0x0f, r8
  mov.b r5, r15
  rla.b r15
  rla.b r15
  rla.b r15
  rla.b r15
  rla.b r15
  bis.b r15, r7
  bis.b r6, r8
  mov.b r7, &DCOCTL
  mov.b r8, &BCSCTL1
  ret

send_settings:
  ;; Send DCO
  mov.w r5, r15
  add.w #hex_table, r15
  mov.b @r15, r15
  call #uart_send_char
  mov.b #',', r15
  call #uart_send_char
  ;; Send RSEL
  mov.w r6, r15
  add.w #hex_table, r15
  mov.b @r15, r15
  call #uart_send_char
  mov.b #',', r15
  call #uart_send_char
  ;; Send CPU Speed (it's in little endian in memory so go backwards)
  mov.w #CPU_FREQ_HI+1, r13
cpu_freq_next_byte:
  mov.b @r13, r15
  call #send_hex
  sub.w #1, r13
  cmp.w #CPU_FREQ_LO, r13
  jhs cpu_freq_next_byte
  mov.b #',', r15
  call #uart_send_char
  ;; Send DCO count to verify CPU_FREQ
  mov.w #DCO_COUNT+1, r13
dco_count_next_byte:
  mov.b @r13, r15
  call #send_hex
  sub.w #1, r13
  cmp.w #DCO_COUNT, r13
  jhs dco_count_next_byte
  ;; Send \r\n
  mov.b #'\r', r15
  call #uart_send_char
  mov.b #'\n', r15
  call #uart_send_char
  call #delay
  ret

send_hex:
  mov.b r15, r7
  and.b #0x0f, r7
  rrc.b r15
  rrc.b r15
  rrc.b r15
  rrc.b r15
  and.b #0x0f, r15
  add.w #hex_table, r15
  add.w #hex_table, r7
  mov.b @r15, r15
  call #uart_send_char
  mov.b @r7, r15
  call #uart_send_char
  ret

; send_string(r15)
send_string:
  cmp.b #0, 0(r15)
  jz send_string_exit
send_string_wait_tx:
  bit.b #UCA0TXIFG, &IFG2
  jz send_string_wait_tx
  mov.b @r15+, &UCA0TXBUF
  jmp send_string
send_string_exit:
  ret

; uart_send_char(r15)
uart_send_char:
  bit.b #UCA0TXIFG, &IFG2
  jz uart_send_char
  mov.b r15, &UCA0TXBUF
  ret

wait_on_dco_sample:
  mov.w #0, r9
wait_on_dco_sample_loop:
  cmp.w #4, r9
  jne wait_on_dco_sample_loop
  ret

calibrate_baud_rate:
  ;; 512 = 2^9
  mov.w r11, r14
  mov.w r14, &DCO_COUNT
  mov.w #0, r15
  ;mov.w #4, r13
  mov.w #9, r13
mul_2_to_9:
  add.w r14, r14
  addc.w r15, r15
  dec r13
  jnz mul_2_to_9
  ;; r15:r14 now contains cpu frequency
  mov.w r14, &CPU_FREQ_LO
  mov.w r15, &CPU_FREQ_HI
  ;; 9600 = 0x2580
  ;; 57600 = 0xe100
  mov.w #0x2580, r4
  ;mov.w #0xe100, r4
  mov.w #0, r13
divide_baud:
  add.w #1, r13
  sub.w r4, r14
  subc.w #0, r15
  jn divide_baud_done
  jmp divide_baud
divide_baud_done:
  mov.w r13, &BAUD_DIV
  ;; Setup UART
  mov.b #UCSSEL_2|UCSWRST, &UCA0CTL1
  mov.b #0, &UCA0CTL0
  mov.b &BAUD_DIV, &UCA0BR0
  mov.b &BAUD_DIV+1, &UCA0BR1
  bic.b #UCSWRST, &UCA0CTL1
  ret

delay:
  mov.w #0, r15
delay_loop:
  dec.w r15
  jnz delay_loop
  ret

timer_interrupt_a:
  add.w #1, r10
  reti

timer_interrupt_b:
  add.w #1, r9
  mov.w r10, r11
  mov.w #0, r10
  reti

.align 16
done_string:
.db "done\r\n",0

.align 16
hex_table:
.db "0123456789ABCDEF"

vectors:
.org 0xfff2
  dw timer_interrupt_a     ; Timer0_A3 TACCR0, CCIFG
.org 0xfffa
  dw timer_interrupt_b     ; Timer1_A3 TACCR0, CCIFG
.org 0xfffe
  dw start                 ; Reset



